// Code generated by go-swagger; DO NOT EDIT.

package models

// This file was generated by the swagger tool.
// Editing this file might prove futile when you re-run the swagger generate command

// MmdsContentsObject Describes the contents of MMDS in JSON format.
//
// swagger:model MmdsContentsObject
type MmdsContentsObject interface{}
