//go:build linux && cgo
// +build linux,cgo
/*
 * Copyright 2022 Jim Ramsay <jramsay@redhat.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define KUBENS_ENVNAME "KUBENSMNT" // Note: Must be kept manually in-sync with common.go

char* getKubeNsMnt();
int joinMountNamespace(const char* namespace, char* errorMessage, size_t errlen);
